// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include <cutil.h>

// includes, kernels
#include <saxpy_kernel.cu>

////////////////////////////////////////////////////////////////////////////////
// declaration, forward
void runTest(int argc, char** argv);
void randomInit(float*, int);
void printDiff(float*, float*, int, int);


////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int
main(int argc, char** argv)
{
    runTest(argc, argv);

    CUT_EXIT(argc, argv);
}

////////////////////////////////////////////////////////////////////////////////
//! Run a simple test for CUDA
////////////////////////////////////////////////////////////////////////////////
void
runTest(int argc, char** argv)
{
    CUT_DEVICE_INIT();

    // set seed for rand()
    srand(2006);

    // allocate host memory for matrices x and y
    unsigned int size_x = SIZE;
    unsigned int mem_size_x = sizeof(float) * size_x;
    float* h_x = (float*) malloc(mem_size_x);
    unsigned int size_y = SIZE;
    unsigned int mem_size_y = sizeof(float) * size_y;
    float* h_y = (float*) malloc(mem_size_y);

	float alpha=1.0;
	unsigned int incx=1,incy=1;
    // initialize host memory
    randomInit(h_x, size_x);
    randomInit(h_y, size_y);

    // allocate device memory
    float* d_x;
    CUDA_SAFE_CALL(cudaMalloc((void**) &d_x, mem_size_x));
    float* d_y;
    CUDA_SAFE_CALL(cudaMalloc((void**) &d_y, mem_size_y));

    // copy host memory to device
    CUDA_SAFE_CALL(cudaMemcpy(d_x, h_x, mem_size_x,
                              cudaMemcpyHostToDevice) );
    CUDA_SAFE_CALL(cudaMemcpy(d_y, h_y, mem_size_y,
                              cudaMemcpyHostToDevice) );

int i;
for (i=0;i<BLOCK_NUM*THREAD_NUM;i++)
	printf("%.3f  %.3f\n",h_x[i],h_y[i]);	

    
    // create and start timer
    unsigned int timer = 0;
    CUT_SAFE_CALL(cutCreateTimer(&timer));
    CUT_SAFE_CALL(cutStartTimer(timer));

    // setup execution parameters
    dim3 threads(THREAD_NUM,1);
    dim3 grid(BLOCK_NUM,1);

    // execute the kernel
    saxpy<<< grid, threads >>>(alpha,d_x,incx,d_y,incy);

    // check if kernel execution generated and error
    CUT_CHECK_ERROR("Kernel execution failed");

    // copy result from device to host
    CUDA_SAFE_CALL(cudaMemcpy(h_y, d_y, mem_size_y,
                              cudaMemcpyDeviceToHost) );

    // stop and destroy timer
    CUT_SAFE_CALL(cutStopTimer(timer));
    printf("Processing time: %f (ms) \n", cutGetTimerValue(timer));
    CUT_SAFE_CALL(cutDeleteTimer(timer));

for (i=0;i<BLOCK_NUM*THREAD_NUM;i++)
	printf("%.3f\n",h_y[i]);	

    // clean up memory
    free(h_x);
    free(h_y);

    CUDA_SAFE_CALL(cudaFree(d_x));
    CUDA_SAFE_CALL(cudaFree(d_y));

}

// Allocates a vector with random float entries.
void randomInit(float* data, int size)
{
    for (int i = 0; i < size; ++i)
        data[i] = rand() / (float)RAND_MAX;
}



